.po -3
.nh
.TH "Application" n "automatically Fri Mar  1 10:58:03 1991" "ET++ version v2.2" "ET++ class description"
.SH NAME
Application.short \- generic handling of documents and application commands
.SH DESCRIPTION
The class \fBApplication\fP is the central part of the \fBgeneric ET++ application\fP. It manages a list of documents, parses the command line, opens the application window and handles application commands.
.br

.br
\fBDocument Handling\fP
.br

.br
\fBApplication\fP manages an unlimited number of documents which are stored in the \fIinstvar documents\fP. Every document has a certain type (see also \fIclass Document\fP). By default, the application creates new and opens existing documents of one preferred type which is called the \fBmain document type\fP and is stored in the \fIinstvar mainDocumentType\fP. Documents of other types can be opened by overriding the \fImethod CanOpenDocument\fP and checking for additional types. Regardless whether a new document is created or an existing document is opened the respective type is always passed to the \fImethod DoMakeDocuments\fP which must create a document of that type.
.br

.br
\fBApplication Window\fP
.br

.br
\fBApplication\fP creates and opens the application window. This window is stored in the \fIinstvar applDialog\fP and is an instance of the \fIclass ApplDialog\fP. The default layout of the application window contains the \fBnew, open and quit buttons\fP for creating a new document, opening an existing document or quitting the application. This layout can be customized by overriding the \fImethod DoCreateDialog\fP. The application window is opened at a default position. By overriding the \fImethod OpenApplicationDialog\fP this position can be modified or the opening can be avoided at all.
.br

.br
\fBEvent Handling\fP
.br

.br
\fBApplication\fP is the next handler of its documents and of the application window. The application itself has no next handler. It always returns \fBNULL\fP in the \fImethod GetNextHandler\fP.
.br
Events not processed by a document or the application window are forwarded to the \fImethod DoMenuCommand\fP or the \fImethod Control\fP of the application which, by default, passes them to the \fImethod HandleApplicationCommands\fP. This method processes only some pre-defined events. User-defined commands can be handled by overriding the \fImethod DoMenuCommand\fP or the \fImethod Control\fP. See also \fIcookbook 'Event Handling'\fP.
.br

.br
\fBCommand Line Parsing\fP
.br

.br
At application startup the \fImethod ParseCommandLine\fP scans the command line for arguments starting with a "-". Single options are passed to the \fImethod DoParseOptions\fP which, by default, recognizes the ET++ options (the "-E..." options). Command line arguments not starting with a "-" are interpreted as names of documents to be opened initially. These names are passed to the \fImethod OpenDocument\fP.
.br
The \fImethod DoParseOptions\fP must be overridden if the application has to process additional command line options. If the command line is interpreted in a completely different way, then the \fImethod ParseCommandLine\fP should be overriden.
.br

.br
\fBExample\fP
.br

.br
The class \fBApplication\fP is the starting point for the application developer. He always has to create a subclass and at least override the \fImethod DoMakeDocuments\fP to create new documents of the required type. \fBmain\fP creates an instance of this derived class and starts the application through the \fImethod Run\fP as shown in the following example:
.na
.nf

.na
.nf

#include "ET++.h"
#include "MyDocument.h"

const char *cDocTypeMyType = "MYDOCTYPE";  // main document type

class MyApplication : public Application {
public:
    MyApplication (int argc, char **argv)
                   : Application (argc, argv, cDocTypeMyType)
        { }
    Document *DoMakeDocuments (const char *type)
        { return new MyDocument(type); }
};

main (int argc, char **argv)
{
    MyApplication myapp (argc, argv);
    return myapp.Run();
}


.ad b
.fi

.ad b
.fi


.br
classes are always derived from \fBApplication\fP.
.br
class \fBApplication\fP
is never reused directly.
.br
class \fBApplication\fP is abstract.
.br
class \fBApplication\fP contains 37 methods.
.PD 0
.TP 20
\fBowner of class:\fR
nobody.
.PD 0
.TP 20
\fBbaseclasses:\fR
.na
EvtHandler
.ad b
.PD 0
.TP 20
\fBflags:\fR
EvtFlags

.SH CLASS VARIABLES
.TP
.BR fileDialog " (protected FileDialog *)"
the file dialog window displayed by the \fImethod Open\fP which allows the user to select the name of a document to be opened.
.br
\fBfileDialog\fP is initially \fBNULL\fP. The \fImethod Open\fP initializes it with an instance of the \fIclass FileDialog\fP returned by the \fImethod MakeFileDialog\fP (delayed creation).

.SH INSTANCE VARIABLES
.TP
.BR applDialog " (protected Dialog *)"
the dialog window opened when the application is started. It is commonly referred to as \fBapplication window\fP. \fBapplDialog\fP is initialized in the \fImethod Run\fP with an instance of the \fIclass ApplDialog\fP. By default the window contains only the \fBnew, open and quit buttons\fP. This layout can be customized by overriding the \fImethod DoCreateDialog\fP. After creation the \fImethod OpenApplicationDialog\fP opens this window at a position returned by the \fImethod GetNewDocumentPos\fP. The \fImethod Run\fP closes it immediately before the application terminates.

.TP
.BR argc " (protected int)"
the \fBnumber of command line arguments\fP passed to \fBmain\fP. The program name is also counted so \fBargc\fP is always at least 1. The same value is also accessible through the \fIvar gArgc\fP.

.TP
.BR argv " (protected char **)"
an array of strings with the \fBcommand line arguments\fP passed to \fBmain\fP. It always contains at least one entry which is the program name and can additionally be accessed by the \fIvar gProgname\fP.
.br
At application startup the command line arguments are processed by the \fImethod ParseCommandLine\fP. The command line arguments are also accessible through the \fIvar gArgv\fP.

.TP
.BR clipboard " (protected ClipBoard *)"
refers to the \fBglobal ET++ clipboard\fP which is usually accessed by the \fIvar gClipBoard\fP. The \fImethod Run\fP creates the clipboard at application startup and destroys it before the application terminates.
.br
See \fIvar gClipBoard\fP and \fIclass ClipBoard\fP.

.TP
.BR documents " (protected ObjList *)"
contains the \fBlist of documents\fP opened by the application. \fBdocuments\fP is initialized in the constructor with an instance of the \fIclass ObjList\fP. It is modified by the \fImethod AddDocument\fP and the \fImethod RemoveDocument\fP. Specific documents can be searched with the \fImethod FindDocument\fP.
.br
.B Known Problems:
It is not possible to use another kind of collection. \fBdocuments\fP should have been declared as \fBCollection *\fP.

.TP
.BR gBatch " (protected bool)"
indicates whether drawing on windows is done in \fBdouble buffering mode\fP. See \fIvar gBatch\fP.

.TP
.BR gClassManager " (protected ClassManager *)"
refers to the \fBET++ class manager\fP which is created at application startup.
.br
See \fIvar gClassManager\fP and \fIclass ClassManager\fP.

.TP
.BR gDebug " (protected bool)"
indicates whether the application runs in \fBdebug mode\fP. See \fIvar gDebug\fP.

.TP
.BR gEnviron " (protected char **)"
an array of strings holding the names and values of the \fBenvironment variables\fP of the application. Environment variables are stored in the format "NAME=value".
.br
Use the \fIvar gSystem\fP and the appropriate methods of the \fIclass System\fP to test and modify environment variables.
.br
.B Known Problems:
The Inspector displays only the first environment variable. Maybe the type specified in the \fBMetaImpl\fP statement in "Application.C" is wrong ?!

.TP
.BR gObservers " (protected IdDictionary *)"
refers to the \fBobserver dictionary\fP managed by the \fIclass Object\fP. This dictionary associates objects with a collection of their observers. To deal with observers use the appropriate methods of the \fIclass Object\fP.

.TP
.BR gScreenRect " (protected Rectangle)"
the size of the screen. See \fIvar gScreenRect\fP.

.TP
.BR gWindow " (protected BlankWin *)"
the window which currently has the input focus. See \fIvar gWindow\fP.

.TP
.BR inited " (protected bool)"
indicates whether the application has already been initialized. At application startup the \fImethod ApplInit\fP sets it to \fBTRUE\fP.

.TP
.BR lastDocPos " (protected Point)"
the position where the main window of the last document has been opened. The \fImethod GetNewDocumentPos\fP increments it by a certain value. \fBlastDocPos\fP is used to nicely arrange the windows of the documents.

.TP
.BR mainDocumentType " (protected char *)"
the preferred type of documents the application handles. It is passed to the \fImethod DoMakeDocuments\fP whenever a new document is created by the \fImethod NewDocument\fP. If the user selects an existing document for opening the \fImethod CanOpenDocument\fP compares the type of that document with the \fBmainDocumentType\fP.

.TP
.BR menu " (protected Menu *)"
the popup menu displayed by the application window. The menu is created in the \fImethod GetMenu\fP (delayed creation).

.TP
.BR printHierarchy " (protected bool)"
This flag is set to \fBTRUE\fP if the option \fB-Eh\fP appears on the command line. In this case a table describing the \fBclass hierarchy\fP is printed on standard output.

.TP
.BR version " (protected char *)"
a string containing \fBversion and copyright\fP information which is displayed by the \fImethod About\fP.


.SH INSTANCE METHOD LIST

.PD 0
.TP 20
application commands
About
.br
HandleApplicationCommands
.br
Quit
.br
Run

.PD 0
.TP 20
application window
DoCreateDialog
.br
OpenApplicationDialog
.br
ShowApplicationWindow

.PD 0
.TP 20
command line
DoParseOptions
.br
GetName
.br
ParseCommandLine
.br
ProgramName

.PD 0
.TP 20
constructor
Application

.PD 0
.TP 20
debugging
Debug

.PD 0
.TP 20
destructor
~Application

.PD 0
.TP 20
document handling
AddDocument
.br
CanOpenDocument
.br
CloseAllDocuments
.br
DoMakeDocuments
.br
FindDocument
.br
GetMainDocumentType
.br
GetNewDocumentPos
.br
MakeFileDialog
.br
NewDocument
.br
Open
.br
OpenDocument
.br
RemoveDocument

.PD 0
.TP 20
dynamic loading
DynLoad
.br
DynLoadHook

.PD 0
.TP 20
error handling
DoOnError

.PD 0
.TP 20
initialization
ApplInit

.PD 0
.TP 20
miscellaneous
Control
.br
DoMenuCommand
.br
GetMenu
.br
GetNextHandler
.br
InspectorId
.br
Parts

.PD 0
.TP 20
progenv support
UpdateGlobals

.SH CATEGORIES

.na
Application Framework
.ad b

.SH FILES
.PD 0
.TP 20
declaration:
\fBApplication.h

.SH HISTORY
joe   - Mon Apr  8 11:28:39 MET DST 1991 - created
.br
joe   - Wed May 29 20:18:46 MET DST 1991 - SBG changes
.br
chris - Thu May 30 16:30:00 MET DST 1991 - reviewed
.br
joe   - Mon Jun  3 14:35:35 MET DST 1991 - minor changes
.br
