.po -3
.nh
.TH "Cluster" n "automatically Thu Mar 21 20:11:29 1991" "ET++ version v2.2" "ET++ class description"
.SH NAME
Cluster.short \- arranges VObjects in rows and/or columns
.SH DESCRIPTION
The \fBCluster\fP arranges its contained VObjects according to the align specification in the \fIinstvar align\fP in rows and/or columns and leaves a minimum gap (see \fIinstvar minGap\fP) between the objects.
.br
The flags in the \fIinstvar align\fP (see also \fIenum VObjAlign\fP) control the layout:
.br
If only horizontal or only vertical flags are set then the \fBCluster\fP makes only one column, respectively one row. If both horizontal and vertical flags are set then the \fBCluster\fP tries to make a square layout by setting the number of columns (see \fIinstvar cols\fP) to the square root of the number of inner VObjects.
.br
Examples:
.na
.nf

	only horizontal flags set		only vertical flags set
		1				1  2  3  4
		2
		3
		4

	both horizontal and vertical flags set (square layout)
			1  3
			2  4

.ad b
.fi
The horizontal flags determine whether the VObjects of a column are aligned left, centered or right within that column.  The vertical flags determine whether the VObjects in a row are aligned at the top, centered, at their common baseline or at the bottom of that row.  If the align flags \fBeVObjHExpand\fP or \fBeVObjVExpand\fP are set and there is extra space available for the \fBCluster\fP, then the VObjects are additionally expanded in the respective direction.

.br
classes are rarely derived from \fBCluster\fP.
.br
class \fBCluster\fP
is often reused directly.
.br
class \fBCluster\fP contains 18 methods.
.PD 0
.TP 20
\fBowner of class:\fR
nobody.
.PD 0
.TP 20
\fBbaseclasses:\fR
.na
CompositeVObject
.ad b
.PD 0
.TP 20
\fBsubclasses:\fR
.na
EnumItem, Form, ManyOfCluster, OneOfCluster
.ad b
.PD 0
.TP 20
\fBflags:\fR
CompositeVObjectFlags

.SH INSTANCE VARIABLES
.TP
.BR actGap " (protected Point)"
is the actual computed gap between the contained VObjects. Its minimum value is the \fIinstvar minGap\fP. The \fImethod SetExtent\fP calculates the \fBactGap\fP based on the actually available space for the \fBCluster\fP.
.br
The \fBactGap\fP can be accessed with the \fImethod GetActGap\fP.

.TP
.BR align " (protected VObjAlign)"
contains the alignment specification how to arrange the contained VObjects (see also \fIenum VObjAlign\fP). If only horizontal or only vertical align flags are set then the \fBCluster\fP makes only one column, respectively one row. If both horizontal and vertical flags are set then the \fBCluster\fP makes a square layout. This calculation is done by the \fImethod CalcCols\fP.
.br
\fBalign\fP can be accessed with the \fImethod GetAlign\fP.

.TP
.BR bs " (protected short *)"
is an array containing the baselines of all rows. Thus the allowed index range is \fB0 ... rows-1\fP. The memory for \fBbs\fP is allocated by the \fImethod CalcDimensions\fP and freed by the destructor. The \fImethod CacheMinSize\fP computes the baselines. If the VObjects must be aligned at their baselines (i.e. the flag \fBeVObjVBase\fP of the \fIinstvar align\fP is set), then the baseline of a row is set to the greatest baseline, otherwise to the greatest height of the VObjects in that row. The cached baselines are used by the \fImethod SetOrigin\fP.
.br
See also \fIinstvar ht\fP and \fIinstvar wd\fP.

.TP
.BR cols " (protected short)"
contains the number of columns of the \fBCluster\fP. \fBcols\fP is calculated by the \fImethod CalcCols\fP and set by the \fImethod CalcDimensions\fP.
.br
See also \fIinstvar rows\fP.

.TP
.BR ht " (protected short *)"
is an array containing the heights of all rows. Thus the allowed index range is \fB0 ... rows-1\fP. The memory for \fBht\fP is allocated by the \fImethod CalcDimensions\fP and freed by the destructor. The \fImethod CacheMinSize\fP sets the height of a row to the greatest height of the VObjects in that row. The cached heights are used by the \fImethod SetOrigin\fP and the \fImethod SetExtent\fP.
.br
See also \fIinstvar bs\fP and \fIinstvar wd\fP.

.TP
.BR minGap " (protected Point)"
is the minimum space the \fBCluster\fP reserves between two VObjects. The actual gap, which is calculated by the \fImethod SetExtent\fP and stored in the \fIinstvar actGap\fP, may be greater.
.br
The \fBminGap\fP can be accessed with the \fImethod GetMinGap\fP.

.TP
.BR minsize " (protected Metric)"
is the overall minimum size of the \fBCluster\fP. It is computed by the \fImethod CacheMinSize\fP and cached in this variable for efficiency because the layout calculation is rather expensive, especially when several Clusters are nested.

.TP
.BR rows " (protected short)"
contains the number of rows of the \fBCluster\fP. The value of \fBrows\fP is set in the \fImethod CalcDimensions\fP and automatically results from the previously calculated value of the \fIinstvar cols\fP.

.TP
.BR wd " (protected short *)"
is an array containing the widths of all columns. Thus the allowed index range is \fB0 ... cols-1\fP. The memory for \fBwd\fP is allocated by the \fImethod CalcDimensions\fP and freed by the destructor. The \fImethod CacheMinSize\fP sets the width of a column to the greatest width of the VObjects in that column. The cached widths are used by the \fImethod SetOrigin\fP and the \fImethod SetExtent\fP.
.br
See also \fIinstvar bs\fP and \fIinstvar ht\fP.


.SH INSTANCE METHOD LIST

.PD 0
.TP 20
client interface
Cluster
.br
GetActGap
.br
GetAlign
.br
GetItem
.br
GetMinGap

.PD 0
.TP 20
implementation
Base
.br
CalcDimensions
.br
GetMinSize
.br
Init
.br
PrintOn
.br
ReadFrom
.br
SetExtent
.br
SetOrigin
.br
~Cluster

.PD 0
.TP 20
overrider interface
CacheMinSize
.br
CalcCols

.SH CATEGORIES

.na
VObject Composition
.ad b

.SH FILES
.PD 0
.TP 20
declaration:
\fBCluster.h

.SH KNOWN PROBLEMS

See \fImethod GetItem\fP and \fImethod SetExtent\fP.
.SH HISTORY
joe@csesbg.at	Thu Jun 20 17:51 MET DST 1991	1st version
.br
chris@csesbg.at	Fri Jun 21 16:52:11 MET DST 1991	review, minor changes
.br
