.po -3
.TH "Document::DoWrite" n "automatically Fri Mar 15 14:13:05 1991" "ET++ version v2.2" "ET++ method description"
.SH NAME
Document::DoWrite \- instance method
.SH TEMPLATE
void DoWrite(ostream & outputStream, int saveOption)
.SH SPECIFIERS
public
virtual
.SH DESCRIPTION
\fBDoWrite\fP is called by the \fImethod Store\fP and must \fBwrite the document's data\fP to the \fBoutputStream\fP. The data can be written by means of the \fBObject Input/Output Mechanism\fP (see \fIclass Object\fP and \fIcookbook Object Input/Output\fP).
.br
The overridden method should always call the base class' method first so that \fBDocument::DoWrite\fP can write a special line including the \fIinstvar docType\fP. This allows the \fIclass FileType\fP to determine the type of files. The line looks like this: \fB@!DOCUMENT_TYPE ApplicationName\fP.
.br
An example for implementing this method:
.na
.nf


class MyDocument : public Document {
    Object *op;        // some data
    ...
};

void MyDocument::DoWrite(ostream & os, int saveOption)
{
    Document::DoWrite(os, saveOption);

    os \<\< op;
}


.ad b
.fi
The \fBsaveOption\fP normally has no meaning and contains 0. A different \fBsaveOption\fP is only possible if \fBDoWrite\fP is triggered by the \fImethod SaveAs\fP. \fBSaveAs\fP gets the save option from the save file dialog which is created by the \fImethod MakeFileDialog\fP. This method must be overridden to create a special file dialog allowing the user to specify other save options (see \fIclass FileDialog\fP).
.br
\fBDoWrite\fP can then use the \fBsaveOption\fP to choose a different output format, e.g. to write the data in pure ASCII format instead of the ET++ Object I/O format.
.br
See also the \fBtree\fP example application.
Method is always overridden.
.SH ARGUMENTS
.TP
\fBostream & outputStream
.br
the output stream where to write the document's data
.TP
\fBint saveOption
.br
a user-defined option
.SH RETURN ARGUMENT
void

.SH CATEGORIES

.na
input/output, overrider interface
.ad b

.SH FIRST DEFINITION
class Document
.SH FILES
.PD 0
.TP 20
implementation:
\fBDocument.C

