.po -3
.TH "EvtHandler::DoMenuCommand" n "automatically Fri Mar 15 14:09:27 1991" "ET++ version v2.2" "ET++ method description"
.SH NAME
EvtHandler::DoMenuCommand \- instance method
.SH TEMPLATE
commandToExecute = Command *
DoMenuCommand(int commandId)
.SH SPECIFIERS
public
virtual
.SH DESCRIPTION
\fBDoMenuCommand\fP is called by the \fImethod VObject::DoRightButtonDownCommand\fP when the user selected a command from a popup menu. The default implementation forwards \fBDoMenuCommand\fP to the next handler.
.br
A typical implementation processes certain commands and passes remaining commands to the base classes and next handlers as shown in the following example:
.na
.nf

Command *MyClass::DoMenuCommand (int commandId)
{
    switch (commandId)
    {
        case cMENUCMD1:        // create an undoable command
            return new MenuCmd1(/* ... */);

        case cMENUCMD2:
            PerformMenuCmd2();     // execute menu command directly
            break;

        default:      // pass other commands to base class
            return MyBaseClass::DoMenuCommand(commandId);
    }
    return gNoChanges;
}


.ad b
.fi
If the selected menu command causes changes, then an undoable command should be created. Other commands can be executed directly. See the \fImethod TextView::DoMenuCommand\fP for a concrete example.
.br
See also \fImethod GetMenu\fP, \fImethod DoCreateMenu\fP and \fImethod DoSetupMenu\fP.
Method is often overridden.
.SH ARGUMENTS
.TP
\fBint commandId
.br
the id of the selected menu command
.SH RETURN ARGUMENT
.TP
\fBCommand * commandToExecute
.br
a command to be executed or \fBgNoChanges\fP
.SH CATEGORIES

.na
menus, commands, event handling, overrider interface
.ad b

.SH FIRST DEFINITION
class EvtHandler
.SH FILES
.PD 0
.TP 20
implementation:
\fBEvtHandler.C

