.po -3
.TH "Object::FreeAll" n "automatically Fri Mar 1 10:54:57 1991" "ET++ version v2.2" "ET++ method description"
.SH NAME
Object::FreeAll \- instance method
.SH TEMPLATE
void FreeAll()
.SH SPECIFIERS
public
virtual
.SH DESCRIPTION
FreeAll recursively applies FreeAll to all objects referenced by the
receiver which are not acquainted, and deletes them afterwards. Non-regular
objects are simply deleted. Calling \fBFreeAll\fP results in the freeing
of the object graph originating in the receiver. \fBThe receiver is
no more working after a FreeAll\fP. The result of all subsequents calls to
the receiver is undefined, except for the destructor. Note that the
receiver is not deleted; that has to be done explicitly.
.br
Note that a FreeAll operation never sends change messages.
.br
For example, the \fImethod Collection::FreeAll\fP destroys all objects
contained in the collection. The \fImethod CompositeVObject::FreeAll\fP
deletes all VObjects contained in its list and also the list itself.
.br
See \fIcookbook 'Memory Management'\fP.
Method is often overridden.
.SH ARGUMENTS
none

.SH RETURN ARGUMENT
void

.SH CATEGORIES

.na
destruction, forwarding, client interface, overrider interface
.ad b

.SH FIRST DEFINITION
class Object
.SH FILES
.PD 0
.TP 20
implementation:
\fBObject.C

.SH KNOWN PROBLEMS

(1) Using FreeAll has to be exercised with great care because multiple deletions
of the same object must not occur.
.br
(2) FreeAll should leave the receiver working since it is a public
operation. That would be both feasible and conceptually cleaner. Most uses
of FreeAll, however, are in the context of deleting the receiver
afterwards, so the requirement of leaving the receiver working would lead
to tremendous unefficiencies. Thus, FreeAll plays a role comparable to that
of the destructor.
