.po -3
.TH "Object::IsEqual" n "automatically Fri Mar 1 10:54:57 1991" "ET++ version v2.2" "ET++ method description"
.SH NAME
Object::IsEqual \- instance method
.SH TEMPLATE
objectsAreEqual = bool
IsEqual(ObjPtr anObject)
.SH SPECIFIERS
public
virtual
.SH DESCRIPTION
\fBIsEqual\fP checks whether \fBthis\fP and \fBanObject\fP are
equal.
.br
The default implementation simply compares the memory addresses of the two
objects. This means it checks whether \fBthis\fP and \fBanObject\fP are
identical.
.br
\fBIsEqual\fP \fBshould be overridden\fP if objects are put in
collections (see \fIclass Collection\fP).
.br
The typical implementation first checks by means of the \fImacro
IsKindOf\fP or the \fImethod IsA\fP whether \fBanObject\fP is of the
same class as \fBthis\fP. Then it compares the objects based on the values
of some instance variables. See also \fImethod Compare\fP and \fImethod Hash\fP.
.br
\fBExamples\fP
.br
\fImethod ByteArray::IsEqual\fP, \fImethod ObjInt::IsEqual\fP, \fImethod VObject::IsEqual\fP, \fImethod Collection::IsEqual\fP.
Method is often overridden.
.SH ARGUMENTS
.TP
\fBObjPtr anObject
.br
the object to be compared with the receiver
.SH RETURN ARGUMENT
.TP
\fBbool objectsAreEqual
.br
is \fBTRUE\fP if \fBthis\fP is equal to
\fBanObject\fP
.SH CATEGORIES

.na
comparison, collection support, client
interface, overrider interface
.ad b

.SH FIRST DEFINITION
class Object
.SH FILES
.PD 0
.TP 20
implementation:
\fBObject.C

.SH KNOWN PROBLEMS

1. \fBIsEqual\fP is not guaranteed to be commutative: calling \fBIsEqual\fP for two objects may lead to different results because the objects
can be of different classes with different implementations of this method.
Commutativity is guaranteed only if every implementation of \fBIsEqual\fP
assures that both objects are of the same class (e.g. 
.na
.nf
IsA() ==
anObject-\>IsA()
.ad b
.fi
). In ET++ however, most implementations check the class
with the \fImacro IsKindOf\fP. In this case commutativity is not
guaranteed if the classes of the objects are in a superclass/subclass
relationship and the subclass overrides this method.
.br
2. Sometimes more than one kind of equality is required.
.br
3. Implementing \fBIsEqual\fP and \fImethod Compare\fP very often
requires duplication of code.
