.po -3
.nh
.TH "View" n "automatically Mon Apr  8 20:17:22 1991" "ET++ version v2.2" "ET++ class description"
.SH NAME
View \- abstract drawing surface and clipboard operations
.SH DESCRIPTION
The abstract class \fBView\fP provides a \fBdrawing surface\fP of arbitrary size to render and print the data structures of an application. In addition, \fBView\fP maintains a \fBcurrent selection\fP and exchanges data with other Views via the \fBclipboard\fP.
.br

.br
\fBDocument and Views\fP
.br

.br
Views are usually created in the \fImethod Document::DoMakeWindows\fP. A document may have any number of Views showing the document's data in different representations.
.br
A \fBView\fP is usually not directly installed in a window, rather it is put into a \fBClipper\fP or \fBScroller\fP in order to show only a part of reasonable size and to allow scrolling. By installing a \fBView\fP in a \fBSplitter\fP or in more than one \fBScroller\fP different portions of the \fBView\fP can be displayed simultaneously. For example:
.na
.nf


Window *MyDocument::DoMakeWindows()
{
    View *view1 = new MyView1 (this /* the next handler */, /* data */);

    return new Window (this, Point(600,800), eWinDefault,
               new Expander (cIdNone, eVert, gPoint2,
                   new Scroller (view1),        // show view1 in 2 Scrollers
                   new Scroller (view1),        // simultaneously
                   new Splitter (new MyView2 (this, /* data */)),
                   NULL
               )
           );
}

.ad b
.fi

.br
A document is always the next event handler of its \fBViews\fP. A \fBView\fP processes user input and passes command objects along the event handler chain to the document. The document executes these commands and this way modifies its data. Dividing the work of displaying, input handling and storing the data is a common practice in ET++ applications and conforms to the MVC (Model-View-Controller) concept introduced by SmallTalk.
.br

.br
\fBSelection and Clipboard\fP
.br

.br
\fBView\fP maintains a \fBcurrent selection\fP.  The current selection is a part of the data shown by the \fBView\fP, e.g. a range of text or an element of a list of objects. Commands such as cut, copy and paste are always applied to the current selection. The current selection is additionally drawn highlighted (see Drawing and Selection Highlighting below).
.br
The \fBclipboard\fP is a kind of store where a \fBView\fP can deposit its current selection. Copies of the selection can later be used by the same or by other \fBViews\fP. The cut, copy and paste commands use the clipboard mechanism to exchange data between different Views and even between Views of different applications.
.br
The current selection can be cut or copied to the clipboard and the contents of the clipboard can be pasted into the data structure of the \fBView\fP. Subclasses of \fBView\fP implement \fBcut and copy\fP by overriding the \fImethod HasSelection\fP and the \fImethod SelectionToClipboard\fP. The \fBpaste\fP command can be implemented by overriding the \fImethod CanPaste\fP and the \fImethod PasteData\fP. \fBHasSelection\fP and \fBCanPaste\fP are actually only used to setup the cut, copy and paste menu items. A good example for implementing the clipboard operations can be found in the \fIclass TextView\fP.
.br

.br
\fBDrawing and Selection Highlighting\fP
.br

.br
Drawing and highlighting the current selection is done by overriding the \fImethod Draw\fP. For example, a \fBView\fP showing a list of \fBVObjects\fP would implement \fBDraw\fP like this:
.na
.nf


class MyView : public View {
    Collection *list;          // a list of VObjects
    VObject *selected;         // currently selected VObject of the list
public:
    ....
    void Draw (Rectangle);
};

void MyView::Draw (Rectangle rect)
{
    Iter next(list);
    VObject *vop;
    while (vop = (VObject*) next())
    {
        bool highlight = (vop == selected);
        vop-\>DrawAll(rect, highlight);	// draw highlighted if it is selected
    }
}


.ad b
.fi
The \fImethod Update\fP is called before the actual drawing is done. \fBUpdate\fP gives the possibility to update the \fBView\fP, e.g. to recalculate its layout. See the \fImethod DialogView::Update\fP for an example.

.br
classes are always derived from \fBView\fP.
.br
class \fBView\fP
is never reused directly.
.br
class \fBView\fP is abstract.
.br
class \fBView\fP contains 30 methods.
.PD 0
.TP 20
\fBowner of class:\fR
nobody.
.PD 0
.TP 20
\fBbaseclasses:\fR
.na
VObject
.ad b
.PD 0
.TP 20
\fBsubclasses:\fR
.na
CollectionView, DialogView, StaticTextView
.ad b
.PD 0
.TP 20
\fBflags:\fR
ViewFlags

.SH INSTANCE VARIABLES
.TP
.BR clippers " (private SeqCollection *)"
contains all \fBClippers\fP that show this \fBView\fP. Scrolling, invalidation and drawing methods of the \fBView\fP are not only forwarded to the normal container, but to all \fBClippers\fP in this list. This allows to display a \fBView\fP simultaneously in more than one \fBClipper\fP. The first \fBClipper\fP in this list is identical with the \fIinstvar VObject::container\fP.
.br
\fBclippers\fP is modified by the \fImethod AddToClipper\fP and the \fImethod RemoveFromClipper\fP. It can be accessed with the \fImethod GetClipperList\fP.

.TP
.BR focus " (protected Clipper *)"
contains the \fBClipper\fP which currently has the input focus. \fBfocus\fP refers to the focus \fBClipper\fP only when an event is processed. If \fBfocus\fP is set, then scrolling methods are forwarded only to that \fBClipper\fP instead of passing them to all Clippers in the \fIinstvar clippers\fP. See also the \fImethod Input\fP.

.TP
.BR nexthandler " (private EvtHandler *)"
is the next event handler returned by the \fImethod GetNextHandler\fP. It is specified in the constructor but can also be set with the \fImethod SetNextHandler\fP. The \fBnexthandler\fP usually refers to the \fBDocument\fP that created the \fBView\fP in its \fBDoMakeWindows\fP method.


.SH INSTANCE METHOD LIST

.PD 0
.TP 20
client interface
GetClipperList
.br
GetDocument

.PD 0
.TP 20
clipboard
CanPaste
.br
HasSelection
.br
PasteData
.br
SelectionToClipboard

.PD 0
.TP 20
drawing
DrawAll
.br
ShowInAllClippers

.PD 0
.TP 20
event handling
DoCursorKeyCommand
.br
DoFunctionKeyCommand
.br
Input
.br
SetNextHandler

.PD 0
.TP 20
implementation
AddToClipper
.br
CheckOpen
.br
GetNextHandler
.br
GetViewedRect
.br
RemoveFromClipper
.br
SetExtent
.br
~View

.PD 0
.TP 20
invalidation
InvalidateRect

.PD 0
.TP 20
menus
DoCreateMenu
.br
DoMenuCommand
.br
DoSetupMenu

.PD 0
.TP 20
miscellaneous
InspectorId

.PD 0
.TP 20
overrider interface
Update
.br
View

.PD 0
.TP 20
scrolling
ConstrainScroll
.br
RevealAlign
.br
RevealRect
.br
Scroll

.SH CATEGORIES

.na
Application Framework
.ad b

.SH FILES
.PD 0
.TP 20
declaration:
\fBView.h

.SH KNOWN PROBLEMS

See \fImethod ShowInAllClippers\fP.
.SH HISTORY
joe@csesbg.at	Wed Apr 17 15:25 1991	first version
.br
joe@csesbg.at	Tue Jun 25 19:51 1991	reworked
.br
gil@csesbg.at	Tue Jul 09 13:00 1991	reviewed
.br
chris@csesbg.at	Wed Jul 10 09:00 1991	reviewed
.br
.so mann/View::AddToClipper.n
.so mann/View::CanPaste.n
.so mann/View::CheckOpen.n
.so mann/View::ConstrainScroll.n
.so mann/View::DoCreateMenu.n
.so mann/View::DoCursorKeyCommand.n
.so mann/View::DoFunctionKeyCommand.n
.so mann/View::DoMenuCommand.n
.so mann/View::DoSetupMenu.n
.so mann/View::DrawAll.n
.so mann/View::GetClipperList.n
.so mann/View::GetDocument.n
.so mann/View::GetNextHandler.n
.so mann/View::GetViewedRect.n
.so mann/View::HasSelection.n
.so mann/View::Input.n
.so mann/View::InspectorId.n
.so mann/View::InvalidateRect.n
.so mann/View::PasteData.n
.so mann/View::RemoveFromClipper.n
.so mann/View::RevealAlign.n
.so mann/View::RevealRect.n
.so mann/View::Scroll.n
.so mann/View::SelectionToClipboard.n
.so mann/View::SetExtent.n
.so mann/View::SetNextHandler.n
.so mann/View::ShowInAllClippers.n
.so mann/View::Update.n
.so mann/View::View.n
.so mann/View::~View.n
